/*
   Module .............libbt
   File ...............SimpleControl.c
   Author .............Traveler Hauptman
   Creation Date ......Nov 24, 2002

              Copyright (c) 2002 Barrett Technology, Inc.
                         139 Main Street
                        Kendall/MIT Square
                   Cambridge, MA  02142-1528  USA

                         All rights reserved.

 ***********************************************************************
 *                            DISCLAIMER                               *
 *                                                                     *
 *  This software and related documentation are provided to you on     *
 *  an as is basis and without warranty of any kind.  No warranties,   *
 *  express or implied, including, without limitation, any warranties  *
 *  of merchantability or fitness for a particular purpose are being   *
 *  provided by Barrett Technology, Inc.  In no event shall Barrett    *
 *  Technology, Inc. be liable for any lost development expenses, lost *
 *  lost profits, or any incidental, special, or consequential damage. *
 *                                                                     *
 ***********************************************************************
 *
 *  NOTES:
 *   A simple Torque, PID & Trapezoidal velocity trajectory controller
 *
 *  REVISION HISTORY:
 *  021124 - TH - File created
 *
 *======================================================================*/

/*! \file SimpleControl.c
    \brief A simple Torque, PID & Trapezoidal velocity trajectory controller

    SimpleControl provides a set of basic control algorithms that are often useful
    when controlling robots.
    The use of these functions is straighforward
    -# Create a sturcture of type SimpleCtl
    -# SCinit to get everything in a reasonable state.
    -# SCevaluate inside your read-position set-torque loop.
    -# SCsetmode to switch to the mode you want

    - SCstarttrj to move the PID setpoint along a trapezoidal trajectory

*/
#include <math.h>
#include <sys/neutrino.h>
#include <pthread.h>
#include <errno.h>
#include <syslog.h>
#include "SimpleControl.h"

#define sign(x) (x>=0?1:-1)
/*! Initialize the SimpleCtl structure */
int SCinit(SimpleCtl *sc)
{
    int err;

    sc->mode = 0;
    sc->requested_torque = 0.0;
    sc->kill_torque = 500000;
    sc->last_dt = 1;
    sc->command_torque = 0;
    sc->trj.state = 0;
    sc->trj.acc = .00000001;
    sc->trj.vel = .00000001;
    PIDinit(&(sc->pid),0, 0, 0, 0);
    err = pthread_mutex_init(&(sc->mutex),NULL);
    if (err)
    {
        syslog(LOG_ERR, "Could not initialize  mutex for simple controller.");
    }
}

/*! Evaluate the Simple Controller

  See the documentation for SimpleCtl for an overview.
  
  This function takes as input the current position and the time since the last evaluation.
It returns a torque. 

\param position the measured position
\param dt the time step we want to evaluate with.
\return Returns a torque command.
*/
double SCevaluate(SimpleCtl *sc, double position, double dt)
{
    double newcommand;
    double newtorque;
    double cmptorque;
    int err;

    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCevaluate lock mutex failed: %d", err);
    }

    sc->position = position;
    switch (sc->mode)
    {
        case 0://Idle
            err = pthread_mutex_unlock( &(sc->mutex) );
            if(err != EOK)
            {
                syslog(LOG_ERR, "SCevaluate unlock mutex (idle) failed: %d", err);
            }
            return 0.0;
            break;
        case 1://Torque
            sc->command_torque = sc->requested_torque;
            err = pthread_mutex_unlock( &(sc->mutex) );
            if(err != EOK)
            {
                syslog(LOG_ERR, "SCevaluate unlock mutex (Torque) failed: %d", err);
            }
            return sc->command_torque;
            break;
        case 2://PID
            if (sc->trj.state == TRAJ_RUN)
            {
                newcommand = evaluate_trajectory(&(sc->trj),dt);
                sc->pid.yref = newcommand;
            }
            sc->pid.y = sc->position;
            sc->pid.dt = dt;
            newtorque = PIDcalc(&(sc->pid));
            cmptorque = fabs(newtorque);

            if (cmptorque > sc->kill_torque)
            {
                err = pthread_mutex_unlock( &(sc->mutex) );
                if(err != EOK)
                {
                    syslog(LOG_ERR, "SCevaluate:kill_torque unlock mutex (PID) failed: %d", err);
                }
                SCsetmode(sc,2);
                err = pthread_mutex_lock( &(sc->mutex) );
                if(err != EOK)
                {
                    syslog(LOG_ERR, "SCevaluate:kill_torque lock mutex failed: %d", err);
                }
                newtorque = 0.0;
            }
            sc->command_torque = newtorque;
            err = pthread_mutex_unlock( &(sc->mutex) );
            if(err != EOK)
            {
                syslog(LOG_ERR, "SCevaluate unlock mutex (PID) failed: %d", err);
            }
            return newtorque;
            break;
        default:
            sc->error = 1;
            err = pthread_mutex_unlock( &(sc->mutex) );
            if(err != EOK)
            {
                syslog(LOG_ERR, "SCevaluate unlock mutex (default) failed: %d", err);
            }
            return 0.0;
        }

}



/*! \brief Set the controller mode

  Sets the controller mode. To keep anything from "jumping" unexpectedly, we reset our
  pid controller whenever we switch into pidmode.

*/
int SCsetmode(SimpleCtl *sc, int mode)
{
    int err;
    double tmp;
    
    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCsetmode lock mutex failed: %d", err);
    }
    if (sc->trj.state == TRAJ_RUN) sc->trj.state = TRAJ_DONE;
    sc->pid.y = sc->position;
    PIDreset(&(sc->pid));
    tmp = PIDcalc(&(sc->pid));
    sc->mode = mode;
    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCsetmode unlock mutex failed: %d", err);
    }
    return 0;
}


int SCsettorque(SimpleCtl *sc,double torque)
{
    int err;
    
    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCsettorque lock mutex failed: %d", err);
    }
    sc->requested_torque = torque; //use a buffer since we are not necessarily running from the same thread as evaluatecontroller
    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCsettorque unlock mutex failed: %d", err);
    }
    
    if (sc->mode != 1)
    {
        return -1;
    }
    return 0;
}
int SCsetpid(SimpleCtl *sc,double kp,double kd,double ki, double saturation) //!! not thread safe yet! We may update these in the middle of their use.
{

    int err;
    
    if (sc->trj.state == TRAJ_RUN || sc->mode == 2) //make sure we don't screw with stuff in the middle of a trajectory.
        return -1;
        
    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCsetpid lock mutex failed: %d", err);
    }
    sc->pid.Kp = kp;
    sc->pid.Kd = kd;
    sc->pid.Ki = ki;
    sc->pid.saturation = saturation;
    
    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCsetpid unlock mutex failed: %d", err);
    }
    
    return 0;
}

int SCsetcmd(SimpleCtl *sc,double cmd) //sets pid command position. if in traj mode, stops trajectory
{
    int err;
    
    if (sc->trj.state == TRAJ_RUN)
        return -1;
    
    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCsetcmd lock mutex failed: %d", err);
    }
    sc->pid.yref = cmd;
    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCsetcmd unlock mutex failed: %d", err);
    }
    return 0;
}

int SCsettrjprof(SimpleCtl *sc,double vel,double acc)
{
    int err;
    
    if (sc->trj.state == TRAJ_RUN)
        return -1;
    
    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCsettrjprof lock mutex failed: %d", err);
    }
    sc->trj.vel = fabs(vel);
    sc->trj.acc = fabs(acc);
    
    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCsettrjprof unlock mutex failed: %d", err);
    }
    return 0;

}
int SCstarttrj(SimpleCtl *sc,double end)
{
    int err;
    double here;
    
    if (sc->trj.state == TRAJ_RUN)
        return -1;
        
    err = pthread_mutex_lock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCstarttrj lock mutex failed: %d", err);
    }
    here = sc->pid.yref;
    init_trajectory(&(sc->trj),here,end);

    err = pthread_mutex_unlock( &(sc->mutex) );
    if(err != EOK)
    {
        syslog(LOG_ERR, "SCstarttrj unlock mutex failed: %d", err);
    }
    return 0;
}

/*! Calculates all the variables necessary to set up a trajectory.

\param start the starting position
\param end the destination position


*/
void init_trajectory(trajectory *traj, double start, double end) //assumes that velocity and acceleration have been set to reasonable values
{
    double ax; //acceleration x
    double dist;


    traj->cmd = 0;
    traj->start_point = start;
    traj->end_point = end;

    traj->t1 = (traj->vel / traj->acc);

    dist = traj->end_point - traj->start_point;

    if (dist > 0) traj->sign = 1;
    else traj->sign = -1;

    ax = 0.5*traj->acc*traj->t1*traj->t1;

    dist = fabs(dist);
    traj->end = dist;
    if (ax > dist/2)
        ax = dist/2;
    traj->x1 = ax;
    traj->x2 = dist - ax;
    traj->t2 = (dist-(2*ax)) / traj->vel + traj->t1;
    traj->t = 0;
    traj->state = TRAJ_RUN;
    if (dist == 0.0) traj->state = TRAJ_DONE;
}



void TRJinit(npoint_traj *TRJ,trj_leg_struct *cstate) //assumes that velocity and acceleration have been set to reasonable values
{
    double ax; //acceleration x
    double dist;
    double dv,dx;
    int subtype;

    TRJ->c0.x = cstate->x;
    TRJ->c0.v = cstate->v;
    TRJ->c3.x = TRJ->pl.point[TRJ->current_point].x;
    TRJ->c3.v = TRJ->pl.point[TRJ->current_point].v;

    dv = TRJ->c3.v - TRJ->c0.v;
    dx = TRJ->c3.x - TRJ->c0.x;
    //figure out how many sub-legs we have
    if ((TRJ->c0.v >= 0) && (dx >= 0) || (TRJ->c0.v <= 0) && (dx <= 0))
        subtype = 1;
    if (sign(TRJ->c0.v) != sign(TRJ->c3.v))
        subtype = 2;
    else
        subtype = 3;

    //Can we meet trj with the allowed acceleration?
    //acc = 2*(x3-x0)


}
/*! Calculate next position on the trajectory

evaluate_trajectory generates points on a trapezoidal velocity trajectory. The points accerate with
constant acceleration specified by acc up to a maximum velocity specified by vel. The max velocity
is maintained until approximately the time it will take to decellerate at acc to a velocity of zero.

\param *traj pointer to the trajectory structure
\param dt the time step to take (in the same units as vel and acc)

\return The value returned is the next point on the trajectory after a step of dt.


*/
double evaluate_trajectory(trajectory *traj,double dt)
{
    double remaining_time,newtime,result;

    if (traj->state == TRAJ_RUN)
    {
        traj->t += dt;
        if (traj->cmd < traj->x1)
            traj->cmd = 0.5 * traj->acc * traj->t * traj->t;  //x = 1/2 a t^2
        else if (traj->cmd < traj->x2)
            traj->cmd = traj->x1 + traj->vel * (traj->t - traj->t1);
        else
        {
            /* time to hit zero = sqrt( (target position - current position)*2/a)
                     new position = 1/2 acc * (time to hit zero - dt)^2 */
            remaining_time = sqrt((traj->end - traj->cmd)*2/traj->acc);
            if (dt > remaining_time)
            {
                traj->cmd = traj->end;
                traj->state = TRAJ_DONE;
            }

            newtime = (remaining_time-dt);
            if(newtime <= 0)
            {
                traj->cmd = traj->end;
                traj->state = TRAJ_DONE;
            }
            traj->cmd =  traj->end - 0.5 * traj->acc * newtime * newtime;
        }

    }
    result = (traj->start_point + traj->sign*traj->cmd);

    if (isnan(result))
    {
        traj->cmd = traj->end;
        traj->state = TRAJ_DONE;
        return traj->end_point;
    }
    return (traj->start_point + traj->sign*traj->cmd);
}


/*! Initializes the PIDregulator data structure

The function sets the main parameters of a PID regulator. You
will only have to use this function a single time when you create the variable. Before
you start using PIDcalc, remember to set y and yref.
*/
void PIDinit(PIDregulator *pid,double Kp, double Kd, double Ki, double dt)
{
    pid->Kp = Kp;
    pid->Kd = Kd;
    pid->Ki = Ki;
    pid->dt = dt;
    pid->filter_de = 0;
    pid->se = 0;
    pid->e = 0;
    pid->laste = 0;
    pid->de = 0;
    pid->firsttick = 1;
    pid->y = 0;
    pid->yref = 0;
    pid->lastresult = 0;
    pid->de_filter.order = 2;
    pid->de_filter.cutoffHz = 10;
    InitDifferentiator (&(pid->de_filter), .001);
    pid->se_filter.order = 2;
    pid->se_filter.cutoffHz = 10;
    InitDifferentiator (&(pid->se_filter), .001);
    //syslog(LOG_ERR,"%f,%f",pid->de_filter.d[0],pid->de_filter.d[1]);


}
/** Reset the state parameters of the PID regulator

  Sets the accumulated error to zero, the reference value to the present value, and notifies 
  the algorithm that it is re-starting.
*/
void PIDreset(PIDregulator *pid)
{
    pid->se = 0;
    pid->firsttick = 1;
    pid->yref = pid->y;
}

/** Increments the state by dt and returns the output of the regulator.

PIDcalc updates the PIDregulator state variable by incrementing the time by dt
and then calculates the error between y and yref.

\param *pid A pointer to the pre-allocated and initialized PIDregulator structure.

\return The regulator output for the present dt, y, and yref.
*/
double PIDcalc(PIDregulator *pid)
{
    pid->e = pid->yref - pid->y;

    if (pid->firsttick)
    {
        pid->laste = pid->e;
        pid->firsttick = 0;
    }

    pid->de = (pid->e-pid->laste)/pid->dt;  //Backward euler
    if (pid->filter_de)
      pid->de = Filter(&(pid->de_filter),pid->e);
    pid->laste = pid->e;

    pid->se += pid->e;
    pid->fe = Filter(&(pid->se_filter),pid->se);
    if (pid->saturation != 0)
        if ((fabs(pid->se)*pid->Ki) > pid->saturation)
        pid->se = sign(pid->se)*pid->saturation;

    pid->lastresult = pid->Kp*pid->e+pid->Kd*pid->de+pid->Ki*pid->se;

    return(pid->lastresult); //bz
}

/*****************************************************************/
/* filtered_value = F i l t e r ( unfiltered_value ) */
/*****************************************************************/
double Filter (filter_struct *filt, double xnew)
{
    int       i,idx;
    double      ynew;

    idx = filt->index;

    filt->x[idx] = xnew;
    filt->y[idx] = 0;

    for (i=0; i<=filt->order; i++)
        filt->y[idx] += filt->n[i] * filt->x[ (idx+i+1)%(filt->order+1) ]
    - filt->d[i] * filt->y[ (idx+i+1)%(filt->order+1) ];

    ynew  = filt->y[idx];
    filt->index = (++idx) % (filt->order+1);

    return(ynew);
}

/** A filter that doesn't do anything
  This function creates a filter that multiplies the input by 1
**/
void InitPassthroughFilter (filter_struct *filt)
{
    int i;

    filt->index = 0;
    for(i=0; i<=4; i++)
    {
        filt->x[i] = 0;
        filt->y[i] = 0;
        filt->d[i] = 0;
        filt->n[i] = 0;
    }

    filt->n[0] = 1;

}

/************************************************************/
/*   I n i t D i f f e r e n t i a t o r ( )                    */
/************************************************************/
void InitDifferentiator (filter_struct *filt, double sample_time)
{
    //  Initializes coefficients for differentiator plus first or
    //  second order lowpass filter.

    // You must initialize the following arguments in main:
    //    diff.order, diff.cutoffHz, sample_time

    double woT;
    double temp;
    int i;

    filt->index = 0;
    for(i=0; i<=4; i++)
    {
        filt->x[i] = 0;
        filt->y[i] = 0;
    }

    if(filt->order == 1)
    {
        woT = sample_time*filt->cutoffHz*2.0*PI;
        temp = 1.0/(2+woT);

        filt->d[1] = 0.0;
        filt->d[0] = -(2-woT)*temp;

        filt->n[1] = 2.0*filt->cutoffHz*PI*2.0*temp;
        filt->n[0] = -(filt->n[1]);
    }
    else if(filt->order == 2)
    {
        woT = sample_time*filt->cutoffHz*2.0*PI;
        temp = 1.0/ ( 4.0*(1.0+woT) + woT*woT  );

        filt->d[2] = 0.0;
        filt->d[1] = -(8.0 - 2.0*woT*woT) * temp;
        filt->d[0] = ( 4.0*(1.0-woT) + woT*woT  ) * temp;

        filt->n[2] = 2.0*filt->cutoffHz*PI*2.0*woT * temp;
        filt->n[1] = 0.0;
        filt->n[0] = -(filt->n[2]);
    }
    else
    {
        syslog(LOG_ERR,"Filter order must be 1 or 2.\n");

    }

}


/*======================================================================*
 *                                                                      *
 *             Copyright (c) 2003 Barrett Technology, Inc.              *
 *                        139 Main Street                               *
 *                       Kendall/MIT Square                             *
 *                  Cambridge, MA  02142-1528  USA                      *
 *                                                                      *
 *                        All rights reserved.                          *
 *                                                                      *
 *  ******************************************************************  *
 *                            DISCLAIMER                                *
 *                                                                      *
 *  This software and related documentation are provided to you on      *
 *  an as is basis and without warranty of any kind.  No warranties,    *
 *  express or implied, including, without limitation, any warranties   *
 *  of merchantability or fitness for a particular purpose are being    *
 *  provided by Barrett Technology, Inc.  In no event shall Barrett     *
 *  Technology, Inc. be liable for any lost development expenses, lost  *
 *  lost profits, or any incidental, special, or consequential damage.  *
 *  ******************************************************************  *
 *                                                                      
 * CVS info: $Id$
 * CVS automatic log (prune as desired):
 * $Log$
 *
 *======================================================================*/



