/*======================================================================*
 *  Module .............libbt
 *  File ...............btwam.c
 *  Author .............Traveler Hauptman
 *  Creation Date ......15 Feb 2003
 *                                                                      *
 *  ******************************************************************  *
 *                                                                      *
 *  NOTES:
 *
 *  REVISION HISTORY:
 *                                                                      *
 *======================================================================*/

/** \file btwam.c
    Provides functions for controlling a WAM that has a CAN infrastructure.
    
    btwam assumes that you have a 7 DOF or 4 DOF wam that is being controlled
on a CAN bus. It provides some high-level functions to help people get up and 
running quickly.     
*/

/*==============================*
 * INCLUDES - System Files      *
 *==============================*/
#include <semaphore.h>
#include <syslog.h>
#include <sys/neutrino.h>
#include <inttypes.h>
#include <math.h>

/*==============================*
 * INCLUDES - Project Files     *
 *==============================*/
#include "SimpleControl.h"
#include "btcan.h"
#include "btsystem.h"
#include "btwam.h"
#include "globals.h"
#include "control_loop.h"
#include "doublebuffer.h"
#include "gimbals.h"

/*==============================*
 * GLOBAL file-scope variables  *
 *==============================*/
extern sem_t     timer_semaphore;
extern int shutdown_threads;
wam_struct WAM;

/* Define WAM link parameters in meters */
#define L3 0.558 
#define L4 0.291 
#define L7 0.065 
#define d3 0.04763
#define Gx 0.1524
#define Gy 0.4097

/*==============================*
 * Functions                    *
 *==============================*/
 /** Initialize a WAM

  This function calls the necessary lower-level libbt functions to set
	up the pucks, enumerate them and scan in the wam configuration information.
	
  \retval 0 Completed successfully
  \retval -1 Some error occured, Check syslog.
*/
int InitWAM(char *wamfile)
{
  int cnt,ret;
  
    syslog(LOG_ERR, "starting initialize system", ret);
	ret = InitializeSystem("actuators.dat","buses.dat","motors.dat","pucks.dat");
    syslog(LOG_ERR, "done with initialize system: returned %d", ret);
  if (ret)
  {
    syslog(LOG_ERR, "Failed to initialize system: returned %d", ret);
    return (-1);
  }
  SetEngrUnits(1);
  ret = EnumerateSystem();
  if (ret)
  {
    syslog(LOG_ERR, "Failed to enumerate system: returned %d", ret);
    return (-1);
  }
  //DumpData2Syslog();
	//------------------------
  WAM.act = GetActuators(&(WAM.num_actuators));
  ret = LoadWAM(wamfile);
	if (ret)
  {
    syslog(LOG_ERR, "Failed to load wam config file: returned %d", ret);
    return (-1);
  }
  zero_wam_vector(&WAM.Jpos);
  zero_wam_vector(&WAM.Jtrq);
  WAM.Gcomp = 0;
  WAM.Ocomp = 0;
  for(cnt = 0; cnt < 7; cnt++)
  { 
    SCinit(&(WAM.sc[cnt]));
    syslog(LOG_ERR,"Mode after init was %d",WAM.sc[cnt].mode);
    SCsetpid(&(WAM.sc[cnt]),WAM.Kp.q[cnt],WAM.Kd.q[cnt],WAM.Ki.q[cnt],WAM.saturation.q[cnt]);
    SCsettrjprof(&(WAM.sc[cnt]),WAM.vel.q[cnt],WAM.acc.q[cnt]);
  }
	//-------------------

  return 0;
}

void CloseWAM()
{
  CloseSystem();
}

/** Returns a pointer to the SimpleCtl array used by the WAM functions
*/
SimpleCtl * GetWAMsc()
{
  return WAM.sc;
}


/** Load wam configuration data

\verbatim
wam file format
<A>,<B>,<C>
<motor_position> x7  //matches actuator indexes to motor indexes. See Actangle2Mpos
<Zero0> x7
<StopTorque> x7
<Park0>x7
<Order> x7
<Kp> x7
<Kd> x7
<Ki> x7
<saturation> x7
<Vel> x7
<Acc> x7
\endverbatim
*/
int LoadWAM(char *wamfile)
{
  FILE *in;
  if ((in = fopen(wamfile, "r")) == NULL)
  {
    syslog(LOG_ERR, "Could not open actuator data file: %s", wamfile);
    return -1;
  }
  
  fscanf(in,"%lf, %lf, %lf, %lf",&(WAM.A),&(WAM.B),&(WAM.C),&(WAM.D));
  fscanf(in,"%d, %d, %d, %d, %d, %d, %d",&(WAM.motor_position[0]),&(WAM.motor_position[1]),&(WAM.motor_position[2]),&(WAM.motor_position[3]),&(WAM.motor_position[4]),&(WAM.motor_position[5]),&(WAM.motor_position[6]));
  read_wam_vector(in,&(WAM.zero_offsets));
  read_wam_vector(in,&(WAM.stop_torque));
  read_wam_vector(in,&(WAM.park_location));
  fscanf(in,"%d, %d, %d, %d, %d, %d, %d",&(WAM.zero_order[0]),&(WAM.zero_order[1]),&(WAM.zero_order[2]),&(WAM.zero_order[3]),&(WAM.zero_order[4]),&(WAM.zero_order[5]),&(WAM.zero_order[6]));
  read_wam_vector(in,&(WAM.Kp));
  read_wam_vector(in,&(WAM.Kd));
  read_wam_vector(in,&(WAM.Ki));
  read_wam_vector(in,&(WAM.saturation));
  read_wam_vector(in,&(WAM.vel));
  read_wam_vector(in,&(WAM.acc));
  fclose(in);
  return 0;
}
/** Write WAM configuration information to a file.
*/
void SaveWAM(char *wamfile)
{
  FILE *out;
  if ((out = fopen(wamfile, "w")) == NULL)
  {
    syslog(LOG_ERR, "Could not open actuator data file: %s", wamfile);
    return;
  }
  
  fprintf(out,"%lf, %lf, %lf, %lf\n",(WAM.A),(WAM.B),(WAM.C),WAM.D);
  fprintf(out,"%d, %d, %d, %d, %d, %d, %d\n",(WAM.motor_position[0]),(WAM.motor_position[1]),(WAM.motor_position[2]),(WAM.motor_position[3]),(WAM.motor_position[4]),(WAM.motor_position[5]),(WAM.motor_position[6]));
  write_wam_vector(out,&(WAM.zero_offsets));
  write_wam_vector(out,&(WAM.stop_torque));
  write_wam_vector(out,&(WAM.park_location));
  fprintf(out,"%d, %d, %d, %d, %d, %d, %d\n",(WAM.zero_order[0]),(WAM.zero_order[1]),(WAM.zero_order[2]),(WAM.zero_order[3]),(WAM.zero_order[4]),(WAM.zero_order[5]),(WAM.zero_order[6]));
  write_wam_vector(out,&(WAM.Kp));
  write_wam_vector(out,&(WAM.Kd));
  write_wam_vector(out,&(WAM.Ki));
  write_wam_vector(out,&(WAM.saturation));
  write_wam_vector(out,&(WAM.vel));
  write_wam_vector(out,&(WAM.acc));
  fclose(out);
}
/** Syslog wam configuration information for debugging
*/
void DumpWAM2Syslog()
{
  syslog(LOG_ERR,"Dump of WAM data---------------------------");
  syslog(LOG_ERR,"A:%lf, B:%lf, C:%lf, D:%lf",(WAM.A),(WAM.B),(WAM.C),(WAM.D));
  dump_wam_vector(&(WAM.zero_offsets));
  dump_wam_vector(&(WAM.stop_torque));
  dump_wam_vector(&(WAM.park_location));
  syslog(LOG_ERR,"%d, %d, %d, %d, %d, %d, %d",(WAM.zero_order[0]),(WAM.zero_order[1]),(WAM.zero_order[2]),(WAM.zero_order[3]),(WAM.zero_order[4]),(WAM.zero_order[5]),(WAM.zero_order[6]));
  syslog(LOG_ERR,"%d, %d, %d, %d, %d, %d, %d",(WAM.motor_position[0]),(WAM.motor_position[1]),(WAM.motor_position[2]),(WAM.motor_position[3]),(WAM.motor_position[4]),(WAM.motor_position[5]),(WAM.motor_position[6]));
  dump_wam_vector(&(WAM.Kp));
  dump_wam_vector(&(WAM.Kd));
  dump_wam_vector(&(WAM.Ki));
  dump_wam_vector(&(WAM.saturation));
  dump_wam_vector(&(WAM.vel));
  dump_wam_vector(&(WAM.acc));
  syslog(LOG_ERR,"Num pucks %d",WAM.num_actuators);
  syslog(LOG_ERR,"End dump of WAM data-----------------------------");
}
/** This function closes the control loop on the WAM using a PID regulator in joint space.
*/

void WAMControlThread(void *data)
{
  int cnt;
  wam_vector Gtrq;
  int idx;
  DoubleBuffer_struct *db;
  data_to_log *dat;

  db = (DoubleBuffer_struct *)data;

  zero_wam_vector(&Gtrq);

  while (!shutdown_threads)
  {
    sem_wait(&timer_semaphore);
    
    GetPositions();
    
    ActAngle2Mpos(&(WAM.Mpos)); //Move motor angles into a wam_vector variable
    Mpos2Jpos(&(WAM.Mpos), &(WAM.Jpos)); //Convert from motor angles to joint angles
    
    if (db->Log_Data) //Handle data logging
    {
      dat = GetDLptr(db);
      for (cnt = 0;cnt < 7;cnt++) dat->Pos[cnt] = WAM.Jpos.q[cnt];
      UpdateDL(db);
    }
    
    for (cnt = 0; cnt < WAM.num_actuators; cnt++) //Calculate control torque for each joint.
    {
      idx = WAM.motor_position[cnt];
      WAM.Jtrq.q[idx] = SCevaluate(&(WAM.sc[idx]),WAM.Jpos.q[idx],.002);
    }
    
    if (WAM.Gcomp)  //Calculate Gravity compensation torque
    {
        calcGcomp(&(WAM.Jpos), &Gtrq, WAM.A,WAM.B,WAM.C,WAM.D);
        for (cnt = 0; cnt < WAM.num_actuators; cnt++)
        {
          idx = WAM.motor_position[cnt];
          WAM.Jtrq.q[idx] += Gtrq.q[idx];
        }
    }
    
    Jtrq2Mtrq(&(WAM.Jtrq), &(WAM.Mtrq));  //Convert from joint torques to motor torques
    Mtrq2ActTrq(&(WAM.Mtrq)); //Move motor torques from wam_vector variable into actuator database
     
    SetTorques();
  }
  pthread_exit(NULL);
}



/** Load a wam_vector with the actuator angles.
 ActAngle2Mpos loads a wam_vector with the actuator angles.
  \param *Mpos Pointer to a wam_vector
  \retval 0 Successful
  \retval -1 Unexpected number of actuators
*/
int ActAngle2Mpos(wam_vector *Mpos) //Extracts actuators 0 thru 6 into wam_vector positions
{
  int num_act, cnt;

  num_act = WAM.num_actuators;
  for (cnt = 0; cnt < num_act; cnt++)
  {
    Mpos->q[WAM.motor_position[cnt]] = WAM.act[cnt].angle;
  }

  return 0;
}
/** Load the actuator database torques with the values from a wam_vector.
 Mtrq2ActTrq loads the actuator database torques with the values from a wam_vector.
  \param *Mtrq Pointer to a wam_vector
  \retval 0 Successful
  \retval -1 Unexpected number of actuators
*/
int Mtrq2ActTrq(wam_vector *Mtrq) //Packs wam_vector of torques into actuator array
{
  actuator_struct *act;
  int num_act, cnt;

  act = GetActuators(&num_act);
  for (cnt = 0; cnt < num_act; cnt++)
  {
      act[cnt].torque = Mtrq->q[WAM.motor_position[cnt]];
  }
  
  return 0;
}
//=================================================working code
/** Transform wam_vector Motor positions to Joint positions
*/
void Mpos2Jpos(wam_vector * Mpos, wam_vector * Jpos) //convert motor angle to joint angle
{
  Jpos->q[0] = ( -Mpos->q[0] / mN1);
  Jpos->q[1] = (0.5 * Mpos->q[1] / mN2) - (0.5 * Mpos->q[2] / mN3);
  Jpos->q[2] = ( -0.5 * Mpos->q[1] * mn3 / mN2) - (0.5 * mn3 * Mpos->q[2] / mN3);
  Jpos->q[3] = (Mpos->q[3] / mN4);
  Jpos->q[4] = 0.5 * Mpos->q[4] / mN5 + 0.5 * Mpos->q[5] / mN5;
  Jpos->q[5] = -0.5 * Mpos->q[4] * mn6 / mN5 + 0.5 * mn6 * Mpos->q[5] / mN5;
  Jpos->q[6] = -1 * Mpos->q[6] / mN7;
}
//=================================================working code
/** Transform wam_vector Joint positions to Motor positions
*/
void Jpos2Mpos(wam_vector * Jpos, wam_vector * Mpos) //convert motor angle to joint angle
{
  Mpos->q[0] = ( -Jpos->q[0] * mN1);
  Mpos->q[1] = (Jpos->q[1] * mN2) - (Jpos->q[2] * mN3 / mn3);
  Mpos->q[2] = ( -Jpos->q[1] * mN2) - ( Jpos->q[2] * mN3 / mn3);
  Mpos->q[3] = (Jpos->q[3] * mN4);
  Mpos->q[4] = Jpos->q[4] * mN5 - Jpos->q[5] * mN5 / mn6;
  Mpos->q[5] = Jpos->q[4] * mN5 + Jpos->q[5] * mN5 / mn6;
  Mpos->q[6] = -Jpos->q[6] / mN7;
}
/** Transform wam_vector Joint torques to Motor torques
*/
void Jtrq2Mtrq(wam_vector * Jtrq, wam_vector * Mtrq) //conbert joint torque to motor torque
{
  Mtrq->q[0] = -Jtrq->q[0] / mN1;
  Mtrq->q[1] = Jtrq->q[1] / mN2 - Jtrq->q[2] * mn3 / mN3;
  Mtrq->q[2] = -Jtrq->q[1] / mN2 - Jtrq->q[2] * mn3 / mN3;
  Mtrq->q[3] = Jtrq->q[3] / mN4;
  Mtrq->q[4] = 0.5 * Jtrq->q[4] / mN5 - mn6 * 0.5 * Jtrq->q[5] / mN5;
  Mtrq->q[5] = 0.5 * Jtrq->q[4] / mN5 + mn6 * 0.5 * Jtrq->q[5] / mN5;
  Mtrq->q[6] = -1 * Jtrq->q[6] / mN7;
}
/** Returns 1 if gravity compensation is active
*/
int getGcomp()
{
	return WAM.Gcomp;
}
/** Enable or disable the gravity compensation calculation
*/
void setGcomp(int onoff)
{
  WAM.Gcomp = onoff;
}
/** Toggle the gravity compensation calculation
*/
void toggleGcomp()
{
  WAM.Gcomp = !WAM.Gcomp;
}

void toggleOcomp()
{
  WAM.Ocomp = !WAM.Ocomp;
}
/** Returns a vector of joint torques necessary to cancel the effect of gravity
*/
void calcGcomp(wam_vector * Jpos, wam_vector * Jtrq,double A,double B,double C,double D) //Calculate gravity compensation torques
{

	int axis;
  wam_vector s,c;

	for(axis=0;axis < 7;axis++)
	{
		s.q[axis]=sin(Jpos->q[axis]);
	  c.q[axis]=cos(Jpos->q[axis]);
	}	
	
	Jtrq->q[0]=0.0;
 	Jtrq->q[1]=(-c.q[1]*c.q[2]*c.q[3] + s.q[1]*s.q[3])*A + (c.q[3]*s.q[1]+c.q[1]*c.q[2]*s.q[3])*B - c.q[1]*c.q[2]*C - s.q[1]*D;
	Jtrq->q[2]=(c.q[3]*s.q[1]*s.q[2])*A - s.q[1]*s.q[2]*s.q[3]*B + s.q[1]*s.q[2]*C;
	Jtrq->q[3]=(-c.q[1]*c.q[3]+c.q[2]*s.q[1]*s.q[3])*A + (c.q[1]*s.q[3]+c.q[2]*c.q[3]*s.q[1])*B;
	Jtrq->q[4]=0.0;
  Jtrq->q[5]=0.0;
  Jtrq->q[6]=0.0;

  if (WAM.Ocomp)
  {
    if (c.q[3] > 0)
      Jtrq->q[2]+=-s.q[2]*c.q[3]*WAM.E;
  }
  
}
/** (internal) Gravity compensation debug calculation
*/
void calcGcomp2(wam_vector * t2, wam_vector * t3,wam_vector * t4) //Calculate gravity compensation torques
{

	int axis;
  wam_vector s,c;

	for(axis=0;axis < WAM.num_actuators;axis++)
	{
		s.q[axis]=sin(WAM.Jpos.q[axis]);
	  c.q[axis]=cos(WAM.Jpos.q[axis]);
	}	
	
 	t2->q[0]=(-c.q[1]*c.q[2]*c.q[3] + s.q[1]*s.q[3]) ; 
  t2->q[1]=(c.q[3]*s.q[1]+c.q[1]*c.q[2]*s.q[3]);
  t2->q[2]= -c.q[1]*c.q[2] ;
  t2->q[3]= -s.q[1];
	t3->q[0]= (c.q[3]*s.q[1]*s.q[2]) ;
  t3->q[1]= -s.q[1]*s.q[2]*s.q[3] ;
  t3->q[2]=s.q[1]*s.q[2];
  t3->q[3]=0;//-s.q[2]*c.q[3];
	t4->q[0]=(-c.q[1]*c.q[3]+c.q[2]*s.q[1]*s.q[3]) ;
  t4->q[1]=(c.q[1]*s.q[3]+c.q[2]*c.q[3]*s.q[1]);
  t4->q[2]=0;
  t4->q[3]=0;
	
}
void setGcompValues(double A,double B,double C,double D)
{
  WAM.A = A;
  WAM.B = B;
  WAM.C = C;
  WAM.D = D;
}
void getGcompValues(double *A,double *B,double *C,double *D)
{
  *A = WAM.A;
  *B = WAM.B;
  *C = WAM.C;
  *D = WAM.D;
}
void getOcompValues(double *E,double *F)
{
  *E = WAM.E;
  *F = WAM.F;
}

void GCompSample(wam_vector *res, double p1, double p2, double p3, double p4)
{
  wam_vector pos[1];
  wam_vector t2,t3,t4;
  
  
  set_wam_vector(pos,p1,p2,p3,p4,0,0,0);
  MoveWAM(pos, &WAM.vel,&WAM.acc);
  sleep(4);
  res->q[0] = WAM.Jtrq.q[0];
  res->q[1] = WAM.Jtrq.q[1];
  res->q[2] = WAM.Jtrq.q[2];
  res->q[3] = WAM.Jtrq.q[3];
}

void GCompFileDump(FILE *out)
{
  wam_vector t2,t3,t4;
  
  calcGcomp2(&t2,&t3,&t4);
  fprintf(out,"%f, %f, %f, %f, %f\n",WAM.Jtrq.q[1],t2.q[0],t2.q[1],t2.q[2],t2.q[3]);
  fprintf(out,"%f, %f, %f, %f, %f\n",WAM.Jtrq.q[2],t3.q[0],t3.q[1],t3.q[2],t3.q[3]);
  fprintf(out,"%f, %f, %f, %f, %f\n",WAM.Jtrq.q[3],t4.q[0],t4.q[1],t4.q[2],t4.q[3]);

}

void GCompCollectData(const char *filename,int j2,int j3, int j4)
{
  FILE *out;
  double octs[5] = {-1.5708,-.785398,0,0.785398,1.5708};
  int cnt1,cnt2,cnt3;
  double pi = 3.14159265;
  
  wam_vector trq[1];
  
  
  out = fopen(filename,"w");
  
  for (cnt1 = 0;cnt1<=j2;cnt1++)
    for (cnt2 = 0;cnt2<=j3;cnt2++)
      for (cnt3 = 0;cnt3<=j4;cnt3++)
    {
      GCompSample(trq,0,-pi/2+cnt1*pi/j2,-pi/2+cnt2*pi/j3,cnt3*pi/j4/2);
      GCompFileDump(out);
    }
  
  fclose(out);
}

/** Take measurements to estimate the coefficients for calculating the effect of gravity
*/
void getGcompvalues()
{
  int cnt,cnt2,idx;
  double t41,t21,t42,t22,t23,t33,t24,t44,t25,t35,te,tf,a,b,c,d;
  wam_vector pos[1],trq[1];

  PowerWAM();
  
  GCompSample(trq,0,-1.5708,0,0);
  t41 = trq->q[3];
  t21 = trq->q[1];
  
  GCompSample(trq,0,-1.5708,0,1.5708);
  t42 = trq->q[3];
  t22 = trq->q[1];
  
  GCompSample(trq,0,-1.5708,-1.5708,1.5708);
  t23 = trq->q[1];
  t33 = trq->q[2];
  
  WAM.A = -t42;
  WAM.B = -t41;
  WAM.C = t33 + WAM.B;
  WAM.D = t21 - t41;
//  WAM.Gcomp = 1;


  SaveWAM("gcomp.dat");
  
}

/** Activate pucks and PID controller
*/
void PowerWAM()
{
  int cnt;
  
  for (cnt = 0;cnt < WAM.num_actuators;cnt++)
  {
    //SCsetmode(&(WAM.sc[WAM.motor_position[cnt]]),0);
    SCsetmode(&(WAM.sc[WAM.motor_position[cnt]]),2);
    SetProp(cnt,CONTROL_MODE,MODE_TORQUE);
    
    
    
  }  
}
/** Deactivate pucks and PID controller
*/
void IdleWAM()
{
  int cnt;
  for (cnt = 0;cnt < WAM.num_actuators;cnt++)
  {
    SetProp(cnt,CONTROL_MODE,MODE_IDLE);
  }  
  for (cnt = 0;cnt < WAM.num_actuators;cnt++)
  {
    SCsetmode(&(WAM.sc[WAM.motor_position[cnt]]),0);
  
  }  
}
/** Reset the position sensors on the pucks to the passed in position in units of joint-space radians
*/
void SetWAMpos(wam_vector *wv)
{
  int cnt;
  long tmp;
  wam_vector motor_angle;
  double result;
  //convert from joint space to motor space, then from motor radians to encoder counts
  
  Jpos2Mpos(wv,&motor_angle);
  for (cnt = 0;cnt < WAM.num_actuators;cnt++)
  {
    result = WAM.act[cnt].motor.counts_per_rev / TWOPI / motor_angle.q[WAM.motor_position[cnt]];
    tmp = floor(result);
    SetProp(cnt,ACTUAL_POSITION,tmp);
    SCevaluate(&(WAM.sc[WAM.motor_position[cnt]]),wv->q[cnt],0);
    SCsetmode(&(WAM.sc[WAM.motor_position[cnt]]),WAM.sc[WAM.motor_position[cnt]].mode);
  }
}
/** Perform a coordinated move of the wam
*/
void MoveWAM(wam_vector *pos, wam_vector *vel, wam_vector *acc)
{
  int cnt,ctr,idx,done = 0,count =0;

  for (cnt = 0;cnt < WAM.num_actuators;cnt++)
  {
    SCsettrjprof(&(WAM.sc[WAM.motor_position[cnt]]),vel->q[WAM.motor_position[cnt]],acc->q[WAM.motor_position[cnt]]);
    SCstarttrj(&(WAM.sc[WAM.motor_position[cnt]]),pos->q[WAM.motor_position[cnt]]);
  }
  while (!done)
  {
    count++;
    ctr = 0;
    for (cnt = 0;cnt < WAM.num_actuators;cnt++)
    { 
      ctr+=WAM.sc[WAM.motor_position[cnt]].trj.state;
    }
    if (ctr == 0) done = 1;
    usleep(1000);
    //if ((count % 1000) == 0)  syslog(LOG_ERR,"waited 1 second to reach position");
  }
}

/** Find the center of a joint range and move to the zero_offsets position relative to the center

\bug Add code to check the puck for IsZerod so that we
know whether we have to zero or not...
*/
void FindWAMcenter(int idx)  //
{
   double this_stop,that_stop,old_kill_torque;

    
    old_kill_torque = WAM.sc[idx].kill_torque;
    WAM.sc[idx].kill_torque = WAM.stop_torque.q[idx];
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx]/2,WAM.acc.q[idx]/2);
    SCstarttrj(&(WAM.sc[idx]),WAM.Jpos.q[idx] + 12 * (WAM.zero_offsets.q[idx]>=0?-1.0:1.0));
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    that_stop = WAM.Jpos.q[idx];
    SCstarttrj(&(WAM.sc[idx]),WAM.Jpos.q[idx] + 12 * (WAM.zero_offsets.q[idx]>=0?1.0:-1.0));
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    this_stop = WAM.Jpos.q[idx];

    SCstarttrj(&(WAM.sc[idx]),(this_stop + that_stop)/2 + WAM.zero_offsets.q[idx]);
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx],WAM.acc.q[idx]);
    WAM.sc[idx].kill_torque = old_kill_torque;
}

void GentleFindWAMcenter(int idx)  //
{
   double this_stop,that_stop,old_kill_torque;

    
    old_kill_torque = WAM.sc[idx].kill_torque;
    WAM.sc[idx].kill_torque = WAM.stop_torque.q[idx];
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx]/2,WAM.acc.q[idx]/2);
    //find stop
    SCstarttrj(&(WAM.sc[idx]),WAM.Jpos.q[idx] + 12 * (WAM.zero_offsets.q[idx]>=0?-1.0:1.0));
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    //back off
    SCstarttrj(&(WAM.sc[idx]),WAM.Jpos.q[idx] - .2 * (WAM.zero_offsets.q[idx]>=0?-1.0:1.0));
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    //find stop at less torque 
    WAM.sc[idx].kill_torque = WAM.stop_torque.q[idx]/2;
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx]/4,WAM.acc.q[idx]/4);
    SCstarttrj(&(WAM.sc[idx]),WAM.Jpos.q[idx] + 12 * (WAM.zero_offsets.q[idx]>=0?-1.0:1.0));
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    that_stop = WAM.Jpos.q[idx];
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx]/2,WAM.acc.q[idx]/2);
    WAM.sc[idx].kill_torque = WAM.stop_torque.q[idx];
    SCstarttrj(&(WAM.sc[idx]),WAM.Jpos.q[idx] + 12 * (WAM.zero_offsets.q[idx]>=0?1.0:-1.0));
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    SCstarttrj(&(WAM.sc[idx]),WAM.Jpos.q[idx] - .2 * (WAM.zero_offsets.q[idx]>=0?1.0:-1.0));
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    WAM.sc[idx].kill_torque = WAM.stop_torque.q[idx]/2;
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx]/4,WAM.acc.q[idx]/4);
    SCstarttrj(&(WAM.sc[idx]),WAM.Jpos.q[idx] + 12 * (WAM.zero_offsets.q[idx]>=0?1.0:-1.0));
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    this_stop = WAM.Jpos.q[idx];

    WAM.sc[idx].kill_torque = WAM.stop_torque.q[idx];
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx]/2,WAM.acc.q[idx]/2);
    SCstarttrj(&(WAM.sc[idx]),(this_stop + that_stop)/2 + WAM.zero_offsets.q[idx]);
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx],WAM.acc.q[idx]);
    WAM.sc[idx].kill_torque = old_kill_torque;
}

/** Find the stop of a joint range and move to the zero_offsets position relative to the stop

  Move in a negative direction if the offset is positive. Search in a positive direction if the offset is negative
  
\bug Add code to check the puck for IsZerod so that we
know whether we have to zero or not...
*/
void FindWAMstop(int idx)  //
{
  double this_stop,that_stop,old_kill_torque;
  double stop_direction;

    stop_direction = (WAM.zero_offsets.q[idx]>0?-1.0:1.0);
    old_kill_torque = WAM.sc[idx].kill_torque;
    WAM.sc[idx].kill_torque = WAM.stop_torque.q[idx];
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx]/2,WAM.acc.q[idx]/2);
    SCstarttrj(&(WAM.sc[idx]),WAM.Jpos.q[idx] + 12 * stop_direction);
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    that_stop = WAM.Jpos.q[idx];
    
    SCstarttrj(&(WAM.sc[idx]),that_stop + WAM.zero_offsets.q[idx]);
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx],WAM.acc.q[idx]);
    WAM.sc[idx].kill_torque = old_kill_torque;
}
void GentleFindWAMstop(int idx)  //
{
  double this_stop,that_stop,old_kill_torque;
  double stop_direction;

    stop_direction = (WAM.zero_offsets.q[idx]>0?-1.0:1.0);
    old_kill_torque = WAM.sc[idx].kill_torque;
    WAM.sc[idx].kill_torque = WAM.stop_torque.q[idx];
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx]/2,WAM.acc.q[idx]/2);
    SCstarttrj(&(WAM.sc[idx]),WAM.Jpos.q[idx] + 12 * stop_direction);
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    SCstarttrj(&(WAM.sc[idx]),WAM.Jpos.q[idx] - .3 * stop_direction);
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    WAM.sc[idx].kill_torque = WAM.stop_torque.q[idx]/2;
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx]/4,WAM.acc.q[idx]/4);
    SCstarttrj(&(WAM.sc[idx]),WAM.Jpos.q[idx] + 12 * stop_direction);
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    that_stop = WAM.Jpos.q[idx];
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx]/2,WAM.acc.q[idx]/2);
    WAM.sc[idx].kill_torque = WAM.stop_torque.q[idx];
    SCstarttrj(&(WAM.sc[idx]),that_stop + WAM.zero_offsets.q[idx]);
    while (WAM.sc[idx].trj.state)
    {
      usleep(1000);
    }
    usleep(500000);
    SCsettrjprof(&(WAM.sc[idx]),WAM.vel.q[idx],WAM.acc.q[idx]);
    WAM.sc[idx].kill_torque = old_kill_torque;
}
/** Apply ZeroWAMaxis to all joints and then set that position to zero.
*/
void ZeroWAM()
{
  int cnt,cnt2,idx;
  double this_stop,that_stop,old_kill_torque;
  wam_vector stops,zero;

  zero_wam_vector(&zero);
  //syslog(LOG_ERR,"zeroWAM A");
  IdleWAM();
  //syslog(LOG_ERR,"zeroWAM B");
  SetWAMpos(&zero);
  //syslog(LOG_ERR,"zeroWAM C");
  PowerWAM();
  //syslog(LOG_ERR,"zeroWAM D");
  /** \bug need to check to see if the actuators we are trying to zero really exist **/
  for (cnt = 0;cnt < WAM.num_actuators;cnt++)
  {
    //syslog(LOG_ERR,"zeroWAM E cnt %d",cnt);
    idx = WAM.zero_order[cnt];    //Validate(idx,0,7,"zerowam");
    
    switch(idx)
    {

        case 1:
        case 2:
        case 3:
        case 6:
        case 7:
        GentleFindWAMcenter(idx-1);
        break;
        
        

        case 4:
        case 5:
        GentleFindWAMstop(idx-1);
        break;
        
        default:
        break;
    }
  }
  
  IdleWAM();
  SetWAMpos(&zero);
  PowerWAM();
}
/** Move the WAM to the park position defined in the wam data file
*/
void ParkWAM()
{
  wam_vector tmpvel,tmpacc;
  
  copy_wam_vector(&tmpvel,&(WAM.vel));
  scale_wam_vector(&tmpvel,0.2);
  copy_wam_vector(&tmpacc,&(WAM.acc));
  scale_wam_vector(&tmpacc,0.2);
  MoveWAM(&(WAM.park_location),&tmpvel,&tmpacc);
}


/** Fills a wam_vector with zeros
*/
void zero_wam_vector(wam_vector *wv)
{
  int cnt;
  
  for(cnt = 0; cnt < 7; cnt++)
    wv->q[cnt] = 0;
}
/** Scalar multiplication of a wam_vector
*/
void scale_wam_vector(wam_vector *wv, double a)
{
  int cnt;
  
  for(cnt = 0; cnt < 7; cnt++)
    wv->q[cnt] *= a;
}
/** Initialize a wam_vector to the given numbers
*/
void set_wam_vector(wam_vector *dest, double s0, double s1, double s2, double s3, double s4, double s5, double s6)
{
  dest->q[0] = s0;
  dest->q[1] = s1;
  dest->q[2] = s2;
  dest->q[3] = s3;
  dest->q[4] = s4;
  dest->q[5] = s5;
  dest->q[6] = s6;
}
/** Copy the contents of the source wam_vector to the destination one
*/
void copy_wam_vector(wam_vector *dest, wam_vector *src)
{
  int cnt;
  
  for(cnt = 0; cnt < 7; cnt++)
    dest->q[cnt] = src->q[cnt];
}
/** fscanf a wam_vector from the file pointed to by FILE *in
*/
int read_wam_vector(FILE *in,wam_vector *wv)
{
  int ret;
  zero_wam_vector(wv);
  ret = fscanf(in,"%lf, %lf, %lf, %lf, %lf, %lf, %lf",&(wv->q[0]),&(wv->q[1]),&(wv->q[2]),&(wv->q[3]),&(wv->q[4]),&(wv->q[5]),&(wv->q[6]));
  if (ret == EOF)
    syslog(LOG_ERR,"read wam vector returned EOF");
  return ret;
}

/** fprintf a wam_vector to FILE *out
*/
int write_wam_vector(FILE *out,wam_vector *wv)
{

  fprintf(out,"%lf, %lf, %lf, %lf, %lf, %lf, %lf\n",(wv->q[0]),(wv->q[1]),(wv->q[2]),(wv->q[3]),(wv->q[4]),(wv->q[5]),(wv->q[6]));
  return 0;
}
/** Syslog a wam_vector
*/
int dump_wam_vector(wam_vector *wv)
{
  syslog(LOG_ERR,"%lf, %lf, %lf, %lf, %lf, %lf, %lf",(wv->q[0]),(wv->q[1]),(wv->q[2]),(wv->q[3]),(wv->q[4]),(wv->q[5]),(wv->q[6]));
  return 0;
}
/** Copies the internal state (position & torque of motors & joints) to user provided wam_vector
*/
void getWAMjoints(wam_vector *Mpos,wam_vector *Mtrq,wam_vector *Jpos,wam_vector *Jtrq)
{
  copy_wam_vector(Mpos,&WAM.Mpos); 
  copy_wam_vector(Mtrq,&WAM.Mtrq); 
  copy_wam_vector(Jpos,&WAM.Jpos); 
  copy_wam_vector(Jtrq,&WAM.Jtrq ); 
}
/** Get data the matches actuator index to wam motor index
*/
void getWAMmotor_position(int *mp)
{
  int cnt;
  for(cnt = 0; cnt < 7; cnt++)
  {
    mp[cnt] = WAM.motor_position[cnt];
  }
}

/** Compute the translation and rotation from the base  
    frame to the endpoint of the WAM arm.  The rotation matrix
    theFrame.rotMatrix[row][column] is calculated, but left to
    the user to interpret using any method they choose.
    
    0      [ r r r x ]
     T   = [ r r r y ]
      tip  [ r r r z ]
           [ 0 0 0 1 ]
*/
int whereAmI(frame *theFrame, int wamType) 
{ 
    double c1, c2, c3, c4, c5, c6, c7; /* Cosines of the joint angles */
    double s1, s2, s3, s4, s5, s6, s7; /* Sines of the joint angles */
    double gimbals[3];
    
    /* Calculate the cosines of the joint angles */
    c1 = cos(WAM.Jpos.q[0]); 
    c2 = cos(WAM.Jpos.q[1]); 
    c3 = cos(WAM.Jpos.q[2]); 
    c4 = cos(WAM.Jpos.q[3]); 
    
    /* Calculate the sines of the joint angles */
    s1 = sin(WAM.Jpos.q[0]); 
    s2 = sin(WAM.Jpos.q[1]); 
    s3 = sin(WAM.Jpos.q[2]); 
    s4 = sin(WAM.Jpos.q[3]); 
    
    /* Depending of the type of WAM, perform the proper kinematics calculation */
    switch(wamType)
    {
        case WAM_4DOF: /* 4 Degree-Of-Freedom WAM */
        /* This is the cartesian translation from the world frame to the end point */
        theFrame->x = -c1*c2*c3*c4*d3 + c1*c2*c3*d3 + c1*c2*c3*s4*L4 + c1*s2*c4*L4 + c1*s2*d3*s4 + c1*s2*L3 + c4*d3*s1*s3 - s1*s3*d3 - s1*s4*L4*s3; 
        theFrame->y = c1*s3*d3 + c1*s3*L4*s4 - c1*c4*d3*s3 - s1*c2*c3*c4*d3 +  s1*c2*c3*d3 + s1*c2*c3*s4*L4 + s1*s2*c4*L4 + s1*s2*d3*s4 + s1*s2*L3; 
        theFrame->z = c2*c4*L4 + c2*d3*s4 + c2*L3 + c3*c4*d3*s2 - c3*s2*d3 - c3*s2*L4*s4; 
        
        /* This is the end frame rotation matrix: rotMatrix[row][column] */
        theFrame->rotMatrix[0][0] = 0;
        theFrame->rotMatrix[0][1] = 0;
        theFrame->rotMatrix[0][2] = 0;
        
        theFrame->rotMatrix[1][0] = 0;
        theFrame->rotMatrix[1][1] = 0;
        theFrame->rotMatrix[1][2] = 0;
        
        theFrame->rotMatrix[2][0] = 0;
        theFrame->rotMatrix[2][1] = 0;
        theFrame->rotMatrix[2][2] = 0;
        break;
        
        case WAM_4DOF_G: /* 4 Degree-Of-Freedom WAM + Gimbals */

        getGimbalsAngles(gimbals);
        
        c5 = cos(gimbals[0]); 
        c6 = cos(gimbals[1]); 
        c7 = cos(gimbals[2]);
        
        s5 = sin(gimbals[0]); 
        s6 = sin(gimbals[1]); 
        s7 = sin(gimbals[2]); 

        /* This is the cartesian translation from the world frame to the end point */
/******************************************************************************************/
        
        theFrame->x = ((c4*Gx+d3)*c3*c1+c1*c3*s4*Gy)*c2+(c1*c4*Gy-c1*s4*Gx+c1*L3)*s2+(-s4*Gy-c4*Gx-d3)*s3*s1;
        theFrame->y = ((c4*Gx+d3)*c1+c1*s4*Gy)*s3+(((c4*Gx+d3)*c3+c3*s4*Gy)*c2+(c4*Gy-s4*Gx+L3)*s2)*s1;
        theFrame->z = (c4*Gy-s4*Gx+L3)*c2+((-c4*Gx-d3)*c3-c3*s4*Gy)*s2;
        
        /* This is the end frame rotation matrix: rotMatrix[row][column] */
        theFrame->rotMatrix[0][0] = (c1*c3*c4*s6*c7+c1*c3*s4*c5*s7+c1*s5*s7*s3)*c2+(-c1*s4*s6*c7+c1*c4*c5*s7)*s2+(c3*s5*s7+(-c4*s6*c7-s4*c5*s7)*s3)*s1+((c1*c3*s4*s5-c1*s3*c5)*c2+c1*c4*s5*s2+(-s3*s4*s5-c3*c5)*s1)*c7*c6; 
        theFrame->rotMatrix[0][1] = (-c1*c3*c4*s6*s7+c1*c3*s4*c5*c7+c1*s5*c7*s3)*c2+(c1*s4*s6*s7+c1*c4*c5*c7)*s2+(c3*s5*c7+(c4*s6*s7-s4*c5*c7)*s3)*s1-((c1*c3*s4*s5-c1*s3*c5)*c2+c1*c4*s5*s2+(-s3*s4*s5-c3*c5)*s1)*s7*c6;
        theFrame->rotMatrix[0][2] = -((c1*c3*s4*s5-c1*s3*c5)*c2+c1*c4*s5*s2+(-s3*s4*s5-c3*c5)*s1)*s6+(c1*c2*c3*c4-c1*s2*s4-s3*c4*s1)*c6; 
        
        theFrame->rotMatrix[1][0] = -c1*c3*s5*s7+(c1*c4*s6*c7+c1*s4*c5*s7)*s3+((c3*c4*s6*c7+c3*s4*c5*s7+s5*s7*s3)*c2+(-s4*s6*c7+c4*c5*s7)*s2)*s1+(c1*s3*s4*s5+c1*c3*c5+((c3*s4*s5-s3*c5)*c2+c4*s5*s2)*s1)*c7*c6;
        theFrame->rotMatrix[1][1] = -c1*c3*s5*c7+(-c1*c4*s6*s7+c1*s4*c5*c7)*s3+((-c3*c4*s6*s7+c3*s4*c5*c7+s5*c7*s3)*c2+(s4*s6*s7+c4*c5*c7)*s2)*s1-(c1*s3*s4*s5+c1*c3*c5+((c3*s4*s5-s3*c5)*c2+c4*s5*s2)*s1)*s7*c6;
        theFrame->rotMatrix[1][2] = -(c1*s3*s4*s5+(s2*c4*s5-c2*s3*c5)*s1+(s1*c2*s4*s5+c1*c5)*c3)*s6+(c1*s3*c4-s1*s2*s4+s1*c2*c4*c3)*c6;
        
        theFrame->rotMatrix[2][0] = (-s4*s6*c7+c4*c5*s7)*c2+(-c3*c4*s6*c7-c3*s4*c5*s7-s5*s7*s3)*s2+(c2*c4*s5+(-c3*s4*s5+s3*c5)*s2)*c7*c6;
        theFrame->rotMatrix[2][1] = (s4*s6*s7+c4*c5*c7)*c2+(c3*c4*s6*s7-c3*s4*c5*c7-s5*c7*s3)*s2-(c2*c4*s5+(-c3*s4*s5+s3*c5)*s2)*s7*c6;
        theFrame->rotMatrix[2][2] = -(c2*c4*s5+(-c3*s4*s5+s3*c5)*s2)*s6+(-s2*c3*c4-c2*s4)*c6;
        break;
/*********************************************************************************************/        
        case WAM_7DOF: /* 7 Degree-Of-Freedom WAM */
        
        c5 = 0; 
        c6 = 0; 
        c7 = 0;
        
        s5 = 0; 
        s6 = 0; 
        s7 = 0; 

        /* This is the cartesian translation from the world frame to the end point */
        theFrame->x = 0;
        theFrame->y = 0;
        theFrame->z = 0;
        
        /* This is the end frame rotation matrix: rotMatrix[row][column] */
        theFrame->rotMatrix[0][0] = 0;
        theFrame->rotMatrix[0][1] = 0;
        theFrame->rotMatrix[0][2] = 0;
        
        theFrame->rotMatrix[1][0] = 0;
        theFrame->rotMatrix[1][1] = 0;
        theFrame->rotMatrix[1][2] = 0;
        
        theFrame->rotMatrix[2][0] = 0;
        theFrame->rotMatrix[2][1] = 0;
        theFrame->rotMatrix[2][2] = 0;        
        break;
        
        default: /* Unknown configuration */
        return(1); /* Return an error */
        break;
    }
    return(0); /* Return SUCCESS */
} 

/*======================================================================*
 *                                                                      *
 *             Copyright (c) 2003 Barrett Technology, Inc.              *
 *                        139 Main Street                               *
 *                       Kendall/MIT Square                             *
 *                  Cambridge, MA  02142-1528  USA                      *
 *                                                                      *
 *                        All rights reserved.                          *
 *                                                                      *
 *  ******************************************************************  *
 *                            DISCLAIMER                                *
 *                                                                      *
 *  This software and related documentation are provided to you on      *
 *  an as is basis and without warranty of any kind.  No warranties,    *
 *  express or implied, including, without limitation, any warranties   *
 *  of merchantability or fitness for a particular purpose are being    *
 *  provided by Barrett Technology, Inc.  In no event shall Barrett     *
 *  Technology, Inc. be liable for any lost development expenses, lost  *
 *  lost profits, or any incidental, special, or consequential damage.  *
 *  ******************************************************************  *
 *                                                                      
 * CVS info: $Id$
 * CVS automatic log (prune as desired):
 * $Log$
 *
 *======================================================================*/
